#include <QtGui>
#include <QSocketNotifier>
#include <QMessageBox>

#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "dialog.h"
#include "clrbox.h"

Dialog::Dialog()
{
	QDialogButtonBox *buttonBox = new QDialogButtonBox(QDialogButtonBox::Close);
	connect(buttonBox, SIGNAL(rejected()), this, SLOT(reject()));

	QGridLayout *lo = new QGridLayout;

	for(int i = 0; i < NCHAN/2; i++) {
		for(int j = 0; j < NCHAN/2; j++)
			lo->addWidget(boxes[i][j] = new clrbox(), i+1, j+1);
	}

	for(int i = 0; i < NCHAN/2; i++)
		lo->addWidget(col[i] = new clrbox(), 0, i+1);
	for(int i = 0; i < NCHAN/2; i++)
		lo->addWidget(row[i] = new clrbox(), i+1, 0);

	QVBoxLayout *mainLayout = new QVBoxLayout;
	mainLayout->addLayout(lo);
	mainLayout->addWidget(buttonBox);
	setLayout(mainLayout);

	setWindowTitle("Capsensor");

	//int fd = ::open("/dev/ttyUSB0", O_RDONLY|O_NONBLOCK, 0644);
	int fd = ::open("myfifo", O_RDONLY|O_NONBLOCK, 0644);
	sn = new QSocketNotifier(fd, QSocketNotifier::Read);
	connect(sn, SIGNAL(activated(int)), this, SLOT(serdata(int)));

	is_syncing = true;
	for(int i = 0; i < NCHAN; i++) {
		fmin[i] = INT_MAX;
		fmax[i] = INT_MIN;
	}
	memset(ch_val, 0, sizeof(ch_val));
	chan = 0;
	skip = 0;
	bytecnt = 0;
}

void Dialog::serdata(int fd)
{
	unsigned char buf[128];
	while(1) {
		ssize_t err = ::read(fd, buf, sizeof(buf));
		if(-1 == err) {
			if(EINTR == errno)
				continue;
			if(EAGAIN == errno)
				return;
			sn->setEnabled(false);
			QMessageBox::critical(this, "Oops", "read failed");
		}
		if(0 == err) {
			sn->setEnabled(false);
			QMessageBox::information(this, "Oops", "Serial died");
		}

		for(int i = 0; i < err; i++)
			handlechar(buf[i]);

		if((size_t)err <= (size_t)sizeof(buf))
			break;
	}
}

void Dialog::handlechar(unsigned ch)
{
	static int xxx = 0;
	if(skip) {
		skip--;
		return;
	}

	if(is_syncing) {
		if(chan != (ch & 0x0f)) {
			chan = 0;
		} else {
			chan++;
			if(chan == NCHAN) {
				is_syncing = false;
				bytecnt = 0;
				fprintf(stderr, "Ok sync\n");
			}
			skip = 2;
		}
	} else {
		c[bytecnt++] = ch;
		if(bytecnt == 3) {
			bytecnt = 0;
			chan = c[0] & 0x0f;
			int cnts = (c[0] << 4) + (c[1] & 0xff);
			int frac = (c[2] & 0xff);
			int v = (int)((int64_t)cnts * 30000000LL / (int64_t)((1<<15) + 3 - frac));
			update_channel(chan, v);
			if(chan < NCHAN/2) {
				col[chan]->setcolor(QColor::fromHsvF(ch_val[chan], 1.0, 1.0-0.8*ch_val[chan]));
			} else {
				row[chan-NCHAN/2]->setcolor(QColor::fromHsvF(ch_val[chan], 1.0, 1.0-0.8*ch_val[chan]));
			}
			if(chan == NCHAN-1) {
				for(int xx = 0; xx < NCHAN/2; xx++) {
					for(int yy = 0; yy < NCHAN/2; yy++) {
						boxes[yy][xx]->setcolor(get_color(xx, yy));
					}
				}
				if(++xxx == 5) {
					repaint();
					xxx = 0;
				}
			}
		}
	}
}

#define DECAY	8000
/* Update channel c with input frequency f */
void Dialog::update_channel(int c, int f)
{
	float d;
	int delta = fmax[c] - fmin[c];
	if(delta > DECAY) {
		fmin[c] += DECAY/10;
		fmax[c] -= DECAY/500;
	}
	if(f < fmin[c])
		fmin[c] = f;
	if(f > fmax[c])
		fmax[c] = f;

	if(fmin[c] == fmax[c])
		d = 1.0;
	else
		d = (float)(f - fmin[c]) / (float)(fmax[c] - fmin[c]);
	ch_val[c] = (ch_val[c] * 2.0 + d * d) / 3.0;
	//ch_val[c] = d * d;
}

/* Get color for position x,y */
QColor Dialog::get_color(int x, int y)
{
	float d = sqrt((ch_val[x] + ch_val[NCHAN/2 + y]) / 2.0);
	return QColor::fromHsvF(d, 1.0, 1.0-0.8*d);
}
